package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.revrobotics.CANSparkMax;
import com.revrobotics.RelativeEncoder;
import com.revrobotics.CANSparkBase.IdleMode;
import com.revrobotics.CANSparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.units.Distance;
import edu.wpi.first.units.Measure;
import edu.wpi.first.units.MutableMeasure;
import edu.wpi.first.units.Units;
import edu.wpi.first.units.Velocity;
import edu.wpi.first.units.Voltage;
import edu.wpi.first.wpilibj.RobotController;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import edu.wpi.first.wpilibj2.command.sysid.SysIdRoutine;
import frc.robot.constants.IntakeConstants;

public class Intake extends SubsystemBase{
    private PIDController intakeAnglePID;

    private CANSparkMax intakeRoller;
    private CANSparkMax intakePivot;

    private RelativeEncoder intakeEncoder;

    private ArmFeedforward intakeFeedForward;

    private MutableMeasure<Voltage> appliedVoltage;
    private MutableMeasure<Distance> distance;
    private MutableMeasure<Velocity<Distance>> velocity;

    private SysIdRoutine intakePivotRoutine;

    public Intake(){
        intakeRoller = new CANSparkMax(IntakeConstants.kIntakeRollerID, MotorType.kBrushless);
        
        intakeRoller.setSmartCurrentLimit(60);
        intakeRoller.setIdleMode(IdleMode.kCoast);
        intakeRoller.burnFlash();

        intakePivot = new CANSparkMax(IntakeConstants.kIntakePivotID, MotorType.kBrushless);

        intakePivot.setIdleMode(IdleMode.kBrake);
        intakePivot.setSmartCurrentLimit(IntakeConstants.kPivotCurrentLimit);
        intakePivot.burnFlash();

        intakeFeedForward = new ArmFeedforward(
            IntakeConstants.kSFeedForward, 
            IntakeConstants.kGFeedForward, 
            IntakeConstants.kVFeedForward
        );

        intakeEncoder = intakePivot.getEncoder();
        intakeEncoder.setPosition(IntakeConstants.kStartingAngle);
        intakeEncoder.setPositionConversionFactor(IntakeConstants.kIntakePivotConversionFactor);

        intakeAnglePID = new PIDController(
            IntakeConstants.kPIntake, 
            IntakeConstants.kIIntake, 
            IntakeConstants.kDIntake
        );

        appliedVoltage = MutableMeasure.mutable(Units.Volts.of(0));
        distance = MutableMeasure.mutable(Units.Meters.of(0));
        velocity = MutableMeasure.mutable(Units.MetersPerSecond.of(0));

        intakePivotRoutine = new SysIdRoutine(
            new SysIdRoutine.Config(),
            new SysIdRoutine.Mechanism(
                (Measure<Voltage> volts) -> {
                    intakePivot.setVoltage(volts.in(Units.Volts));
                }, 
                (log) -> {
                    log.motor("intakePivot")
                        .voltage(appliedVoltage.mut_replace(
                            intakePivot.get() * RobotController.getBatteryVoltage(), Units.Volts
                        ))
                        .linearPosition(
                            distance.mut_replace(intakeEncoder.getPosition(), Units.Meters)
                        )
                        .linearVelocity(
                            velocity.mut_replace(intakeEncoder.getVelocity(), Units.MetersPerSecond)
                        );
                }, 
                this
            )
        );
    }

    public Command intakeDownCommand() {
        return run(() -> {
            intakeRoller.set(1.0);
        
            intakePivot.setVoltage(
                intakeAnglePID.calculate(
                    intakeEncoder.getPosition(), 
                    IntakeConstants.kDownAngle
                ) + intakeFeedForward.calculate(
                    IntakeConstants.kDownAngle, 
                    intakeEncoder.getVelocity()
                )
            );
        });
    }

    public Command manualPivot(DoubleSupplier pivotPower){
        return run(() ->{
            intakePivot.set(pivotPower.getAsDouble());
        });
    }

    public Command climbingStateCommand() {
        return run(() -> {
            intakeRoller.set(0.0);

            intakePivot.setVoltage(
                intakeAnglePID.calculate(
                    intakeEncoder.getPosition(), 
                    IntakeConstants.kDownAngle
                ) + intakeFeedForward.calculate(
                    IntakeConstants.kDownAngle, 
                    intakeEncoder.getVelocity()
                )
            );
        });
    }

    public Command intakeUpCommand() {
        return run(() -> {
            intakeRoller.set(0.0);

            intakePivot.setVoltage(
                intakeAnglePID.calculate(
                    intakeEncoder.getPosition(), 
                    IntakeConstants.kUpAngle
                ) + intakeFeedForward.calculate(
                    IntakeConstants.kUpAngle, 
                    intakeEncoder.getVelocity()
                )
            );
        });
    }

    public Command sysIdQuasistatic(SysIdRoutine.Direction direction) {
        return intakePivotRoutine.quasistatic(direction);
    }

    public Command sysIdDynamic(SysIdRoutine.Direction direction) {
        return intakePivotRoutine.dynamic(direction);
    }
}
