package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.revrobotics.CANSparkMax;
import com.revrobotics.RelativeEncoder;
import com.revrobotics.CANSparkBase.IdleMode;
import com.revrobotics.CANSparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.math.controller.SimpleMotorFeedforward;
import edu.wpi.first.units.Distance;
import edu.wpi.first.units.Measure;
import edu.wpi.first.units.MutableMeasure;
import edu.wpi.first.units.Units;
import edu.wpi.first.units.Velocity;
import edu.wpi.first.units.Voltage;
import edu.wpi.first.wpilibj.DigitalInput;
import edu.wpi.first.wpilibj.Encoder;
import edu.wpi.first.wpilibj.RobotController;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import edu.wpi.first.wpilibj2.command.sysid.SysIdRoutine;
import frc.robot.constants.ShooterConstants;

public class Shooter extends SubsystemBase {
    public enum ShooterSysIDMode {
        PIVOT,
        TOP,
        BOTTOM
    }

    private CANSparkMax topShooter;
    private CANSparkMax bottomShooter;

    private RelativeEncoder bottomEncoder;
    private RelativeEncoder topEncoder;

    private CANSparkMax shooterPivot;

    private Encoder pivotEncoder;

    private PIDController topShooterPID;
    private SimpleMotorFeedforward topShooterFF;

    private PIDController bottomShooterPID;
    private SimpleMotorFeedforward bottomShooterFF;

    private PIDController shooterPivotPID;
    private ArmFeedforward shooterPivotFF;

    private DigitalInput shooterBeamBreak;

    private MutableMeasure<Voltage> appliedVoltage;
    private MutableMeasure<Distance> distance;
    private MutableMeasure<Velocity<Distance>> velocity;

    private SysIdRoutine shooterPivotRoutine;
    private SysIdRoutine shooterTopRoutine;
    private SysIdRoutine shooterBottomRoutine;

    public Shooter(){
        topShooter = new CANSparkMax(ShooterConstants.kTopShooterID, MotorType.kBrushless);
        bottomShooter = new CANSparkMax(ShooterConstants.kBottomShooterID, MotorType.kBrushless);
        
        topShooter.setSmartCurrentLimit(40);
        topShooter.burnFlash();

        bottomShooter.setSmartCurrentLimit(40);
        bottomShooter.burnFlash();

        topEncoder = topShooter.getEncoder();
        bottomEncoder = bottomShooter.getEncoder();

        shooterPivot = new CANSparkMax(ShooterConstants.kShooterPivotID, MotorType.kBrushless);

        shooterPivot.setSmartCurrentLimit(50);
        shooterPivot.burnFlash();

        pivotEncoder = new Encoder(0, 1);
        pivotEncoder.setDistancePerPulse(ShooterConstants.kPivotConversion);

        shooterBeamBreak = new DigitalInput(ShooterConstants.kShooterBeamBreakChannel);

        //TODO These can probably devolve into BangBang controllers and let the feed forwards maintain speed
        topShooterPID = new PIDController(
            ShooterConstants.kShooterP, 
            ShooterConstants.kShooterI, 
            ShooterConstants.kShooterD
        );
        bottomShooterPID = new PIDController(
            ShooterConstants.kShooterP, 
            ShooterConstants.kShooterI, 
            ShooterConstants.kShooterD
        );

        topShooterFF = new SimpleMotorFeedforward(ShooterConstants.kSShooterFF, ShooterConstants.kVShooterFF);
        topShooterFF = new SimpleMotorFeedforward(ShooterConstants.kSShooterFF, ShooterConstants.kVShooterFF);

        shooterPivotPID = new PIDController(
            ShooterConstants.kShooterPivotP,
            ShooterConstants.kShooterPivotI,
            ShooterConstants.kShooterPivotD
        );

        shooterPivotFF = new ArmFeedforward(
            ShooterConstants.kSShooterPivotFF,
            ShooterConstants.kGShooterPivotFF, 
            ShooterConstants.kVShooterPivotFF
        );

        appliedVoltage = MutableMeasure.mutable(Units.Volts.of(0));
        distance = MutableMeasure.mutable(Units.Meters.of(0));
        velocity = MutableMeasure.mutable(Units.MetersPerSecond.of(0));

        shooterPivotRoutine = new SysIdRoutine(
            new SysIdRoutine.Config(), 
            new SysIdRoutine.Mechanism(
                (Measure<Voltage> volts) -> {
                    shooterPivot.setVoltage(volts.in(Units.Volts));
                }, 
                (log) -> {
                    log.motor("shooterPivot")
                        .voltage(appliedVoltage.mut_replace(
                            shooterPivot.get() * RobotController.getBatteryVoltage(), Units.Volts
                        ))
                        .linearPosition(
                            distance.mut_replace(pivotEncoder.getDistance(), Units.Meters)
                        )
                        .linearVelocity(
                            velocity.mut_replace(pivotEncoder.getRate(), Units.MetersPerSecond)
                        );
                }, 
                this
            )
        );

        shooterTopRoutine = new SysIdRoutine(
            new SysIdRoutine.Config(), 
            new SysIdRoutine.Mechanism(
                (Measure<Voltage> volts) -> {
                    topShooter.setVoltage(volts.in(Units.Volts));
                }, 
                (log) -> {
                    log.motor("shooterTop")
                        .voltage(appliedVoltage.mut_replace(
                            topShooter.get() * RobotController.getBatteryVoltage(), Units.Volts
                        ))
                        .linearPosition(
                            distance.mut_replace(topEncoder.getPosition(), Units.Meters)
                        )
                        .linearVelocity(
                            velocity.mut_replace(topEncoder.getVelocity(), Units.MetersPerSecond)
                        );
                }, 
                this
            )
        );

        shooterBottomRoutine = new SysIdRoutine(
            new SysIdRoutine.Config(), 
            new SysIdRoutine.Mechanism(
                (Measure<Voltage> volts) -> {
                    bottomShooter.setVoltage(volts.in(Units.Volts));
                }, 
                (log) -> {
                    log.motor("shooterBottom")
                        .voltage(appliedVoltage.mut_replace(
                            bottomShooter.get() * RobotController.getBatteryVoltage(), Units.Volts
                        ))
                        .linearPosition(
                            distance.mut_replace(bottomEncoder.getPosition(), Units.Meters)
                        )
                        .linearVelocity(
                            velocity.mut_replace(bottomEncoder.getVelocity(), Units.MetersPerSecond)
                        );
                }, 
                this
            )
        );
    }

    public Command angleSpeedsSetpoints(double setpointAngle, double topRPM, double bottomRPM){
        return run(()-> {
            shooterPivot.setIdleMode(IdleMode.kBrake);

            shooterPivot.setVoltage(
                shooterPivotPID.calculate(pivotEncoder.getDistance(), setpointAngle) + 
                shooterPivotFF.calculate(setpointAngle, 0.0));

            topShooter.setVoltage(topShooterPID.calculate(topEncoder.getVelocity(), topRPM)+topShooterFF.calculate(topRPM));

            bottomShooter.setVoltage(bottomShooterPID.calculate(bottomEncoder.getVelocity(), bottomRPM)+bottomShooterFF.calculate(bottomRPM));
        });
    }

    public Command recieveNote(){
        return run(() -> {
            shooterPivot.setIdleMode(IdleMode.kBrake);

            shooterPivot.setVoltage(
                shooterPivotPID.calculate(pivotEncoder.getDistance(), ShooterConstants.kShooterLoadAngle) + 
                shooterPivotFF.calculate(ShooterConstants.kShooterLoadAngle, 0.0));
            
            if(shooterBeamBreak.get()){
                topShooter.set(0.25);
                bottomShooter.set(0.25);
            }else{
                topShooter.set(0.);
                bottomShooter.set(0.); 
            }
        });
    }

    public Command climbState(){
        return run(() -> {
            shooterPivot.setIdleMode(IdleMode.kCoast);
            shooterPivot.set(0.0);
        });
    }

    public DoubleSupplier getShooterAngle(){
        return () -> {return pivotEncoder.getDistance();};
    }

    public Command manualPivot(DoubleSupplier pivotPower){
        return run(() ->{
            shooterPivot.set(pivotPower.getAsDouble());
        });
    }

    public Command sysIdQuasistatic(ShooterSysIDMode mode, SysIdRoutine.Direction direction) {
        switch(mode) {
            case PIVOT:
                shooterPivotRoutine.quasistatic(direction);
                break;
            case TOP:
                shooterTopRoutine.quasistatic(direction);
                break;
            case BOTTOM:
            default:
                shooterBottomRoutine.quasistatic(direction);
                break;
        }

        //It'll never get here, because the switch above has a default case
        return null;
    }

    public Command sysIdDynamic(ShooterSysIDMode mode, SysIdRoutine.Direction direction) {
        switch(mode) {
            case PIVOT:
                shooterPivotRoutine.dynamic(direction);
                break;
            case TOP:
                shooterTopRoutine.dynamic(direction);
                break;
            case BOTTOM:
            default:
                shooterBottomRoutine.dynamic(direction);
                break;
        }

        //It'll never get here, because the switch above has a default case
        return null;
    }
}
